import random


class RandomWord:
    def __init__(self, template: str):
        self.template = template
        self.choices = []
        if "|" in self.template:
            self._parse()
        else:
            self.choices.append(self.template)

    def _parse(self):
        word_begins = 0
        opens = 0
        for i, symbol in enumerate(self.template):
            if symbol == ">":
                opens += 1
            if symbol == "<":
                opens -= 1
            if symbol == "|" and opens == 0:
                self.choices.append(RandomSentence(self.template[word_begins:i]))
                word_begins = i + 1
        if word_begins < len(self.template):
            self.choices.append(RandomSentence(self.template[word_begins:]))

    def generate(self):
        choice = random.choice(self.choices)
        if isinstance(choice, RandomSentence):
            return choice.generate()
        return choice


class RandomSentence(object):
    def __init__(self, template: str, *args, **kwargs):
        self.template = template
        self.kwargs = kwargs
        self.words = []
        self._parse_template()

    def _parse_template(self):
        word = ""
        opens = 0
        for i, symbol in enumerate(self.template):
            if symbol not in (">", "<") or opens != 0:
                word += symbol
            if symbol == "<":
                if opens == 0:
                    self.words.append(RandomWord(word))
                    word = ""
                opens += 1
            if symbol == ">":
                opens -= 1
                if opens == 0:
                    self.words.append(RandomWord(word[:-1]))
                    word = ""
        if word:
            self.words.append(RandomWord(word))

    def generate(self):
        res = "".join(choicer.generate() for choicer in self.words)
        if self.kwargs:
            return res.format(**self.kwargs)
        return res

    def __str__(self):
        return self.generate()

    def __repr__(self):
        return self.generate()



from random import randint

from scintillant.textutils import Phrase


class RandomPhrase(Phrase):
    """ Generator of random phrases.
    :template: str = '{name}, hello!'
    :**kwargs: dict of lists = {
        'name': ['Niel', 'Almaz']
    }
    """

    def get_keyword(self) -> list:
        keywords = []
        for i, val in enumerate(self.template):
            if val == '{':
                for k, sval in enumerate(self.template[i+1:]):
                    if sval == '}':
                        keywords.append(self.template[i+1:i+k+1])
                        break
        return [keyword for keyword in keywords if keyword]

    def __str__(self):
        keywords = self.get_keyword()
        positions = {}
        for word in keywords:
            if word not in self.keywords:
                raise KeyError(f"{word} not found in {self.keywords.keys()} arguments.")
            positions[word] = self.keywords[word][randint(0, len(self.keywords[word]) - 1)]
        print(positions)
        return self.template.format(*self.words, **positions)



"""
Commands for activating the test environment


"""
import sys
import os
import json
import time
from http import HTTPStatus
from zipfile import ZipFile
import multiprocessing
from livereload import Server
import click
import requests

from scintillant import addons_versions
from scintillant.apimodels import User, Client, SkillResponse, SkillRequest, Message
from scintillant.commands.utils import get_config


def create_snlt():
    with open(os.getcwd() + '/.snlt', 'w+') as snlt:
        if sys.platform == 'win32':
            std_path = 'C:\\snlt\\testsuite\\'
        else:
            std_path = '/ust/testsuite/'
        path = input("Specify the path where Python can create "
                     f"the test environment files ({std_path}): ")
        path = std_path if not path else path
        click.echo(path)
        json.dump({
            "skill_name": "bottle-skill-template",
            "skill_working_url": "http://localhost:8080",
            "testsuite": {
                "version": addons_versions['__testsuite__'],
                "path": path,
                "theme": "dark",
                "api_host": "localhost",
                "api_port": "8080"
            }
        }, snlt, sort_keys=True, indent=4)


def update_snlt():
    js = json.load(open(os.getcwd() + '/.snlt'))
    if sys.platform == 'win32':
        std_path = 'C:\\snlt\\testsuite\\'
    else:
        std_path = './testsuite/'
    path = input("Specify the path where Python can create "
                 f"the test environment files ({std_path}): ")
    path = std_path if not path else path
    with open(os.getcwd() + '/.snlt', 'w+') as snlt:
        if 'testsuite' in js and 'path' not in js['testsuite']:
            js['testsuite']['path'] = path
        elif 'testsuite' not in js:
            js['testsuite'] = {'theme': 'dark'}
            js['testsuite']['path'] = path
        json.dump(js, snlt, sort_keys=True, indent=4)


def install_suite(path):
    suite_release = ("https://github.com/PaperDevil/snlt_testsuite/releases/download/"
                     f"{addons_versions['__testsuite__']}"
                     "/dist.zip")
    r = requests.get(suite_release)
    os.makedirs(os.path.dirname(path + 'temp.zip'), exist_ok=True)
    with open(path + 'temp.zip', 'wb') as f:
        f.write(r.content)
    with ZipFile(path + 'temp.zip', 'r') as repo_zip:
        repo_zip.extractall(path)
    os.remove(path + 'temp.zip')


def get_current_app(app_path: str = None):
    try:
        click.echo("Tryng get application instance...")
        sys.path.append(app_path or os.getcwd())
        from manage import run_app
        return run_app
    except ImportError as exc:
        raise ImportError(
            "The specified path does not contain a valid manage.py file. "
            "Make sure you've installed all the dependencies from the "
            "requirements.txt file and created the run_app function correctly "
            "inside the manage.py file!"
        )


def start_bot_server(host: str = 'localhost', port: str = '8080'):
    app = get_current_app()
    click.echo("Starting application server...")
    try:
        app(host=host, port=port)
    except Exception as exc:
        click.echo(exc)


def initial_test_shell(skill_url: str) -> None:
    client: Client = Client(name='testshell')
    user: User = User(idx=0, username='Scintillant', client=client)
    text: str = "!start"
    context: dict = {}
    while True:
        if text in ['!quit', '!exit', '!stop']:
            break
        data = SkillRequest(
            message=Message(user=user, text=text),
            context=context
        ).dict()
        request = requests.post(skill_url + '/skill', json=data).json()
        response = SkillResponse(**request)
        context = response.context
        if response.status == HTTPStatus.RESET_CONTENT:
            click.echo("End of operating...")
            break
        if response.status != HTTPStatus.OK:
            click.echo("End of operating with exception!")
            break

        click.echo(f"{response.status}: {response.text}")
        text = str(click.prompt('', type=str))


def start_test_shell():
    js = get_config()

    if not js:
        create_snlt()
        start_test_shell()
    if 'testsuite' not in js:
        update_snlt()
        start_test_shell()
    # Serving current backend application
    click.echo("Creating server process...")
    _process = multiprocessing.Process(target=start_bot_server, args=(
        js['testsuite'].get('api_host', 'localhost'), js['testsuite'].get('api_port', '8080')
    ))
    _process.start()
    # Starting Shell for testing bot
    click.echo("Starting testshell inputs...")
    time.sleep(3)
    initial_test_shell(js['skill_working_url'])
    _process.terminate()
    _process.join()


def start_test_suite():
    js = get_config()
    if not js:
        create_snlt()
        return start_test_suite()
    if 'testsuite' not in js or 'path' not in js['testsuite']:
        update_snlt()
        return start_test_suite()
    else:
        test_suite_path = js['testsuite']['path']
        install_suite(test_suite_path)

    if not os.path.exists(test_suite_path):
        install_suite(test_suite_path)
        return start_test_suite()

    # Serving current backend application
    click.echo("Creates processes...")
    _process = multiprocessing.Process(target=start_bot_server, args=(
        js['testsuite'].get('api_host', 'localhost'), js['testsuite'].get('api_port', '8080')
    ))
    _process.start()
    # Serving testsuite frontend application
    click.echo("starting testsuite...")
    server = Server()
    server.serve(root=test_suite_path, default_filename='index.html')



""" Default Phrases for user_answers
"""


class Phrase(object):
    def __init__(self, template: str, *args, **kwargs):
        self.template = template
        self.words = args
        self.keywords = kwargs

    def __str__(self):
        return self.template.format(*self.words, **self.keywords)

    def __repr__(self):
        return self.__str__()



import typing
from datetime import datetime
from http import HTTPStatus
from pydantic import BaseModel, Field


class MessageText(str):
    def contains(self, text: str, case_sens=False) -> bool:
        if not case_sens:
            result = text.lower() in self.lower()
        else:
            result = text in self
        return result

    def contains_any(self, texts: list, case_sens=False) -> bool:
        for text in texts:
            if self.contains(text, case_sens):
                return True
        return False


class Client(BaseModel):
    name: str = Field(...)


class User(BaseModel):
    idx: typing.Union[int, str] = Field(..., example='1')
    username: typing.Optional[str] = Field(None, example='ketovx')
    client: typing.Optional[Client]


class Message(BaseModel):
    user: User = Field(...)
    text: MessageText = Field(...)
    binaries: typing.Optional[str] = Field(None)
    timestamp: typing.Union[float, int] = Field(default=datetime.now().timestamp())

    def __init__(self, *args, **kwargs):
        kwargs['text'] = MessageText(kwargs['text'])
        super().__init__(*args, **kwargs)


class SkillResponse(BaseModel):
    """ Response Model for testsuite
    """
    status: HTTPStatus = Field(..., example=HTTPStatus.OK)
    text: typing.Optional[str] = Field(None, example='Any text...')
    binaries: typing.Optional[str] = Field(None)
    context: typing.Optional[dict] = Field(None)


class SkillRequest(BaseModel):
    """ Request model for testsuite
    """
    message: Message
    context: dict = Field({})



"""
scintillant.entry_points.py
~~~~~~~~~~~~~~~~~~~~~~

This module contains the entry-point functions for the scintillant module,
that are referenced in setup.py.
"""
from scintillant import about
from scintillant.commands.templates import start_bottle_skill
from scintillant.commands.testsuite import start_test_suite, start_test_shell
from scintillant.commands.utils import show_testsuite_version

import click


@click.group()
def main() -> None:
    """Main package entry point.

    Delegates to other functions based on user input.
    """
    pass


@main.command()
@click.option('--name', default='bottle-skill-template',
              help="Name of the skill being developed")
def bottle(name):
    """Download the latest skill template"""
    start_bottle_skill(skill_name=name)


@main.command()
def version():
    """Show version of framework"""
    click.echo('Scintillant Framework ' + about['__version__'])
    show_testsuite_version()


@main.command()
def testsuite():
    """Downloads and runs a test environment for the skill."""
    start_test_suite()


@main.command()
def testshell():
    """Start testing shell in current directory"""
    start_test_shell()
